/* sdr.c - 'SCSI Direct Read'
 * Copyright (C) 2000 Tachibana Eriko
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <errno.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/dos.h>
#include <sys/scsi.h>
#include <sys/xglob.h>

#include "sdr.h"


/* File scope function */
static int
get_top_blk (int unit_no, int sec_size, int scsi_id, int size)
{
  unsigned char buf[256 << size];

  /* 擪ZN^ǂݍ */
  if (_scsi_readext (0, 1, scsi_id, size, buf) < 0)
    return -1;

  /* X680x0 */
  if (memcmp ("X68SCSI1", buf, 8) == 0)
    {
      int i;

      /* ItZbg 2048 oCg̃ZN^ǂݍ */
      if (_scsi_readext (2048 >> 8 >> size, 1, scsi_id, size, buf) < 0)
	return -1;
      if (memcmp ("X68K", buf, 4) != 0)
	return -1;

      for (i = 1; i < 16; ++i)
	{
	  if (memcmp ("Human68k", &buf[i * 16 + 0], 8) != 0)
	    continue;			/*  Human68k p[eBV */
	  if (buf[i * 16 + 8] & 1)
	    continue;			/* ANeBuȃp[eBV */

	  if (--unit_no < 0)
	    {
	      int r = (*(int*) &buf[i * 16 + 8]) & 0x00ffffff;

	      return (r << (3 - size)) >> 1;
	    }
	}

      return -1;
    }

#if 0					/* Ή */
  /* IBM-FDISK */
  else if ((buf[0] == 0xb8 || buf[0] == 0xfa || buf[0] == 0x00)
	&& buf[sec_size - 2] == 0x55 && buf[sec_size - 1] == 0xaa)
    {
      return -1;
    }
#endif

  /* IBM-PC Super-Floppy */
  else if (buf[0] == 0xeb
	&& buf[sec_size - 2] == 0x55 && buf[sec_size - 1] == 0xaa)
    {
      return (unit_no == 0) ? 0 : -1;
    }

  /* ZIP Iomega */
  else if (memcmp ("IPAR", buf + 2, 4) == 0)
    {
      return (unit_no == 0) ? 0 : -1;
    }

  return -1;
}


/* Function */
SDR*
SDRopen (const char* filename)
{
  int id, unit_no, sec_size, top_blk, cdrom = 0;
  SDR* sdr;
  SDR* new_sdr;
  int rc;
  int fd;
  union _fcb* fcb;

  if ((sdr = malloc (sizeof (SDR) + 32*1024)) == NULL)
    return NULL;

  if ((fd = _dos_open (filename, 0)) < 0)
    goto dos_error;

  if ((fcb = _dos_getfcb (fd)) < 0)
    goto dos_error2;

  /* SCSI ID ̎ */
  if ((rc = _dos_ioctrlgt (fd)) < 0
   || (rc = (rc & 0x2000) ? _dos_ioctrldvctl (fd, 0x8007, (char*) &id)
			  : _dos_ioctrlrh (fd, (char*) &id, 4)) < 0)
    goto dos_error2;

  {
    int ssp = _dos_super (0);
    unsigned char* dpb = fcb->blk.deventry;

    /* DPB ̎ */
    unit_no = dpb[1];
    sec_size = *(unsigned short*)(dpb + 10);

    /* FAT ̎ */
    if ((rc = _dos_fatchk2 (filename, (unsigned short*) sdr->list - 1, 32*1024)) < 0)
      {
	if (sec_size == 0)
	  {
	    sec_size = 2048;		/* CD-ROM */
	    cdrom = 1;

	    sdr->list[0] = fcb->blk.dirsec;
	    sdr->list[1] = (fcb->blk.size + sec_size - 1) / sec_size;
	    sdr->list[2] = 0;
	    rc = 2 + 4 * 3;
	  }
      }

    if (ssp >=0)
      _dos_super (ssp);
    if (rc < 0)
      goto dos_error2;
  }
  _dos_close (fd);

  if ((new_sdr = realloc (sdr, sizeof (SDR) + rc - 2)) == NULL)
    {
      free (sdr);
      return NULL;
    }
  sdr = new_sdr;

  {
    struct _readcap cap;

    if ((rc = _scsi_readcap (id, &cap)) < 0)
      {
	errno = _errcnv (rc);
	free (sdr);
	return NULL;
      }

    /* ubNTCY 256<<size ̌`ɕϊ */
    switch (cap.size)
      {
      case 256:
	sdr->size = 0;
	break;
      case 512:
	sdr->size = 1;
	break;
      case 1024:
	sdr->size = 2;
	break;
      case 2048:
	sdr->size = 3;
	break;
      default:
	free (sdr);
	errno = EPERM;			/* K */
	return NULL;
      }
  }

  if (!cdrom)
    {
      int* p;

      if ((top_blk = get_top_blk (unit_no, sec_size, id, sdr->size)) < 0)
	{
	  errno = EPERM;		/* K */
	  free (sdr);
	  return NULL;
	}

      for (p = sdr->list; *p;)
	{
	  int n;

	  n = (*p * sec_size) >> 8 >> sdr->size;
	  *p++ = n + top_blk;
	  n = (*p * sec_size) >> 8 >> sdr->size;
	  *p++ = n;
	}
    }

  sdr->id = id;
  sdr->buf_rest = 0;
  sdr->blk_rest = 0;
  sdr->next = sdr->list;

  return sdr;

dos_error2:
  _dos_close (fd);
dos_error:
  free (sdr);
  errno = _errcnv (rc);
  return NULL;
}


/* Function */
int
SDRclose (SDR* sdr)
{
  if (sdr == NULL)
    return -1;

  free (sdr->list);
  free (sdr);
  return 0;
}


/* Function */
int
SDRread (SDR* sdr, void* ptr, int size)
{
  int r = size;
  int blk;

  if (sdr == NULL)
    return 0;

  /* obt@ɎcĂ΍ŏɂo͂ */
  if (sdr->buf_rest)
    {
      int n = min (r, sdr->buf_rest);

      memcpy (ptr, sdr->buf + (256 << sdr->size) - sdr->buf_rest, n);
      sdr->buf_rest -= n;
      ptr += n;
      r -= n;

      if (r == 0)
	return size;
    }

  /* ubNPʂŏ */
  blk = r >> 8 >> sdr->size;
  while (blk)
    {
      int b;

      /* ̃`FC */
      if (sdr->blk_rest == 0)
	{
	  int* p = sdr->next;

	  if (*p == 0)
	    return 0;
	  sdr->blk_no = *p++;
	  sdr->blk_rest = *p++;
	  sdr->next = p;
	}

      /* ̃`FCǂݍރubN */
      b = min (blk, sdr->blk_rest);

      if (_scsi_readext (sdr->blk_no, b, sdr->id, sdr->size, ptr) < 0)
	return 0;
      sdr->blk_no += b;
      sdr->blk_rest -= b;
      ptr += b << 8 << sdr->size;
      blk -= b;
    }

  /* ubN */
  r &= (256 << sdr->size) - 1;
  if (r)
    {
      /* ̃`FC */
      if (sdr->blk_rest == 0)
	{
	  int* p = sdr->next;

	  if (*p == 0)
	    return 0;
	  sdr->blk_no = *p++;
	  sdr->blk_rest = *p++;
	  sdr->next = p;
	}

      /* obt@ɓǂݍŕKvȕo͂ */
      if (_scsi_readext (sdr->blk_no, 1, sdr->id, sdr->size, sdr->buf) < 0)
	return 0;
      ++sdr->blk_no;
      --sdr->blk_rest;

      memcpy (ptr, sdr->buf, r);
      sdr->buf_rest = (256 << sdr->size) - r;
    }

  return size;
}


/* EOF */
